//ExportProc.cpp
#include "export.h"
#include "tinyxml/tinyxml.h"

#include "xmlnode/XmlNode.h"
#include "xmlexp.h"

#ifdef _DEBUG
#pragma comment(lib, "tinyxmld")
#else
#pragma comment(lib, "tinyxml")
#endif


#ifndef		SKELOBJ_CLASS_ID
#define		SKELOBJ_CLASS_ID  	Class_ID( 0x9125, 0 )
#endif

#ifndef		BIPED_CLASS_ID
#define     BIPED_CLASS_ID 		Class_ID( 0x9155, 0 )
#endif

#ifndef		BIPSLAVE_CONTROL_CLASS_ID
#define		BIPSLAVE_CONTROL_CLASS_ID 	Class_ID( 0x9154 , 0 )
#endif

#ifndef		BIPBODY_CONTROL_CLASS_ID
#define		BIPBODY_CONTROL_CLASS_ID 	Class_ID( 0x9156, 0 )
#endif

#ifndef		FOOTPRINT_CLASS_ID
#define		FOOTPRINT_CLASS_ID 	    Class_ID( 0x3011, 0 )
#endif

#ifndef		DUMMY_CLASS_ID
#define		DUMMY_CLASS_ID    Class_ID( 0x876234, 0 )
#endif	


INT_PTR CALLBACK maxProjectOptionsDlgProc(HWND hWnd,UINT message,WPARAM wParam,LPARAM lParam) 
{
	static Export *imp = NULL;

	WPARAM wlParam = LOWORD(wParam);
	WPARAM whParam = HIWORD(wParam);

	switch(message) 
	{
	case WM_INITDIALOG:
		imp = (Export *)lParam;
		CenterWindow(hWnd,GetParent(hWnd));
		return TRUE;

	case WM_CLOSE:
		EndDialog(hWnd, 0);
		return 1;
	case WM_COMMAND:
		{
			if( imp && IDC_EXPORT == wlParam)		
			{
				HWND checkWnd = GetDlgItem( hWnd, IDC_CHECK_LEFTHAND );
				if(SendMessage(checkWnd, BM_GETCHECK, 0, 0) == BST_CHECKED)
					imp->SetLeftHand( true );
				imp->SetExport( true );
				SendMessage(hWnd, WM_CLOSE, 0, 0);				
			}
		}
		return 1;
	}
	return 0;
}
//-----------------------------------------------------------------------------------
XmlExp* Export::WriteBaseMesh( INode* pNode, XmlExp* pParentXML )
{
	XmlExp* pXmlNode = WriteXmlNode( pNode, "GEOMOBJECT_CLASS_ID" );
	pXmlNode->AddAttribute( "name", pNode->GetName() );

	//Object* pObject = pNode->GetObjectRef();
	ObjectState os = pNode->EvalWorldState(0);
	Object* pObject = os.obj;

	//if( pObject->CanConvertToType( Class_ID(TRIOBJ_CLASS_ID, 0)) == 0 )
	//	return pXmlNode;

	//ﰢ ޽ ȯ
	TriObject* pTri = (TriObject *)pObject->ConvertToType(0, Class_ID(TRIOBJ_CLASS_ID, 0));
	if( pTri == NULL )
		return pXmlNode;

	//޽ ü Ѵ.
	Mesh* pMesh	= &pTri->GetMesh();
	if( pMesh == NULL )
		return pXmlNode;

	//޽   ƽ  ߰ ؼ ޽ ͸  Ʒ ߰Ѵ.
	char* szObjName = GetPrimitiveName( pObject );
	XmlNode* pXmlMeshNode = CreateXmlNode<XmlExp>( "mesh" );
	pXmlMeshNode->AddAttribute( "name",szObjName );
	pXmlNode->AttachChild( pXmlMeshNode );

	// 
	int	numVertex	= pMesh->getNumVerts();
	XmlNode* pXmlVertexNode = CreateXmlNode<XmlExp>( "Vertex" );
	pXmlMeshNode->AttachChild( pXmlVertexNode );
	pXmlVertexNode->AddAttribute( "count", ToString( numVertex ).c_str() );

	Matrix3	matWSkin = pNode->GetObjTMAfterWSM(0);  //Matrix3	matWSkin = pNode->GetObjectTM(0);

	for (int i=0; i < numVertex; ++i)
	{
		//Point3 v = pMesh->verts[i];
		Point3 v = matWSkin * pMesh->verts[i];  // ȯ  Ѵ.

		if( GetLeftHand() )
		{
			float temp = v.y;
			v.y = v.z;
			v.z = temp;
		}

		base::Vector3 pos( v.x, v.y, v.z );
		pXmlVertexNode->AttachChild( CreateXmlNode<XmlExp>( "data" ) )->SetData( ToString(pos) );
	}

	// ε
	int numFace = pMesh->getNumFaces();
	XmlNode* pXmlFaceNode = CreateXmlNode<XmlExp>( "Face" );
	pXmlMeshNode->AttachChild( pXmlFaceNode );
	pXmlFaceNode->AddAttribute( "count", ToString( numFace ).c_str() );

	for( int i = 0; i < numFace; ++i )
	{
		base::Index3 index;
		index.a = pMesh->faces[i].v[0];
		index.b = pMesh->faces[i].v[1];
		index.c = pMesh->faces[i].v[2];

		if( GetLeftHand() )
		{
			index.b = pMesh->faces[i].v[2];
			index.c = pMesh->faces[i].v[1];
		}
		pXmlFaceNode->AttachChild( CreateXmlNode<XmlExp>( "data" ) )->SetData( ToString(index) );
	}

	//ְ 
	pMesh->buildNormals();
	XmlNode* pXmlNormalNode = CreateXmlNode<XmlExp>( "Normal" );
	pXmlMeshNode->AttachChild( pXmlNormalNode );
	pXmlNormalNode->AddAttribute( "count", ToString( numVertex ).c_str() );
	for (int i=0; i < numVertex; ++i)
	{
		//ְ 
		Point3 v = Normalize( pMesh->getNormal(i) );	

		if( GetLeftHand() )
		{
			float temp = v.y;
			v.y = v.z;
			v.z = temp;
		}

		base::Vector3 pos( v.x, v.y, v.z );
		pXmlNormalNode->AttachChild( CreateXmlNode<XmlExp>( "data" ) )->SetData( ToString(pos) );
	}

	return pXmlNode;
}
//-----------------------------------------------------------------------------------

#define RetClassName(classId)	if(id == Class_ID( classId, 0 ))	return #classId


char* Export::GetPrimitiveName( Object* pObject )
{
    Class_ID id = pObject->ClassID();

	RetClassName( TRIOBJ_CLASS_ID );
	RetClassName( EDITTRIOBJ_CLASS_ID );
	RetClassName( POLYOBJ_CLASS_ID );
	RetClassName( PATCHOBJ_CLASS_ID );
	RetClassName( NURBSOBJ_CLASS_ID );

	RetClassName( BOXOBJ_CLASS_ID );
	RetClassName( SPHERE_CLASS_ID );
	RetClassName( CYLINDER_CLASS_ID );

    return " ";
}
//-----------------------------------------------------------------------------------
XmlExp* Export::WriteXMLObject( INode* pNode, XmlExp* pParentXML )
{
	//Object* pObject = pNode->GetObjectRef();
	ObjectState os = pNode->EvalWorldState(0);
	Object* pObject = os.obj;

	if(pObject == NULL )
		return NULL;

	SClass_ID lSuperID = pObject->SuperClassID();
	Class_ID lClassID = pObject->ClassID();

	XmlExp* pXmlNode = NULL;
	if(GEOMOBJECT_CLASS_ID == lSuperID )
	{
		if( BONE_OBJ_CLASSID == lClassID || Class_ID(BONE_CLASS_ID, 0) == lClassID)
		{
			pXmlNode = WriteXmlNode( pNode, "BONE_OBJ_CLASSID" );
			pXmlNode->AddAttribute( "name", pNode->GetName() );
		}
		else
		{
			pXmlNode = WriteBaseMesh( pNode, pParentXML );
		}
	}
	else if(CAMERA_CLASS_ID == lSuperID)
	{
		pXmlNode = WriteXmlNode( pNode, "CAMERA_CLASS_ID" );
		pXmlNode->AddAttribute( "name", pNode->GetName() );
	}
	else if(LIGHT_CLASS_ID == lSuperID)
	{
		pXmlNode = WriteXmlNode( pNode, "LIGHT_CLASS_ID" );
		pXmlNode->AddAttribute( "name", pNode->GetName() );
	}
	else if(SHAPE_CLASS_ID == lSuperID)
	{
		pXmlNode = WriteXmlNode( pNode, "SHAPE_CLASS_ID" );
		pXmlNode->AddAttribute( "name", pNode->GetName() );

	}
	else if(HELPER_CLASS_ID == lSuperID)
	{
		pXmlNode = WriteXmlNode( pNode, "HELPER_CLASS_ID" );
		pXmlNode->AddAttribute( "name", pNode->GetName() );
	}

	if( pXmlNode )
		pParentXML->AttachChild( pXmlNode );	

	return pXmlNode;
}
//-----------------------------------------------------------------------------------
XmlExp* Export::WriteAnimation(INode* pNode, XmlExp* pParentXML)
{
#ifdef USE_NOSKIN
	int tickStride = m_aniInfo.tickPerFrame;
	int beginTick = m_aniInfo.beginFrame * tickStride;
	int endTick = m_aniInfo.endFrame * tickStride;
	int aniNum = m_aniInfo.endFrame - m_aniInfo.beginFrame + 1;

	INode*  pParent = pNode->GetParentNode();
	base::Matrix* pAniMat = new base::Matrix[aniNum];

	int i = 0, iTick = beginTick;
	for( ; iTick <= endTick; iTick += tickStride, ++i)
	{
		Matrix3 matLocal;
		Matrix3 matWorld = pNode->GetObjTMAfterWSM( iTick );

		if( pParent == NULL )
		{
			matLocal = matWorld;
		}
		else
		{
			Matrix3 matParent = pParent->GetObjTMAfterWSM( iTick );
			matLocal = matWorld * Inverse( matParent );
		}

		Matrix3 maxMat = matLocal;
		if( GetLeftHand() )
			RightToLeft( &maxMat, &matLocal );

		base::Matrix& mat = pAniMat[i];  mat.Init();	MaxToBaseMatrix( &mat, &maxMat );
	}

	//pAniMat Xml ȯ
	XmlExp* pXmlNode = CreateXmlNode<XmlExp>( "Animation" );
	pXmlNode->AddAttribute( "count", ToString( aniNum ).c_str() );

	pParentXML->AttachChild( pXmlNode );

	for( int i = 0; i < aniNum; ++i)
	{
		base::Matrix& mat = pAniMat[i];
		pXmlNode->AttachChild( CreateXmlNode<XmlExp>( "matrix" ) )->SetData(  ToString( mat ) );
	}
	delete  [] pAniMat;
	return pXmlNode;
#else
	if( IsBone( pNode ) == false )
	{
		pParentXML->AddAttribute( "bone", "false" );
		return NULL;
	}
	pParentXML->AddAttribute( "bone", "true" );

	int tickStride = m_aniInfo.tickPerFrame;
	int beginTick = m_aniInfo.beginFrame * tickStride;
	int endTick = m_aniInfo.endFrame * tickStride;
	int aniNum = m_aniInfo.endFrame - m_aniInfo.beginFrame + 1;

	base::Matrix* pAniMat = new base::Matrix[aniNum];
	Matrix3	matPivot = pNode->GetNodeTM(0);
	matPivot.Invert();

	int i = 0, iTick = beginTick;

	for( ; iTick <= endTick; iTick += tickStride, ++i)
	{
		Matrix3 matWorld = pNode->GetObjTMAfterWSM( iTick );
		matWorld.NoScale();
		
		//ִϸ̼   = Pivot   * ð    
		Matrix3 matAni = matPivot * matWorld;

		Matrix3 maxMat = matAni;
		if( GetLeftHand() )
			RightToLeft( &maxMat, &matAni );

		base::Matrix& mat = pAniMat[i];  mat.Init();	MaxToBaseMatrix( &mat, &maxMat );
	}

	//pAniMat Xml ȯ
	XmlExp* pXmlNode = CreateXmlNode<XmlExp>( "Animation" );
	pXmlNode->AddAttribute( "count", ToString( aniNum ).c_str() );
	pParentXML->AttachChild( pXmlNode );

	for( int i = 0; i < aniNum; ++i)
	{
		base::Matrix& mat = pAniMat[i];
		pXmlNode->AttachChild( CreateXmlNode<XmlExp>( "matrix" ) )->SetData(  ToString( mat ) );
	}
	delete  [] pAniMat;
	return pXmlNode;
#endif
}
//-----------------------------------------------------------------------------------
void Export::WriteHeader( XmlExp* pParentXML )
{
	XmlExp* pXmlNode = CreateXmlNode<XmlExp>( "AnimationHeader" );
	pParentXML->AttachChild( pXmlNode );

	pXmlNode->AttachChild( CreateXmlNode<XmlExp>( "beginFrame" ) )->SetData( ToString( m_aniInfo.beginFrame ) );
	pXmlNode->AttachChild( CreateXmlNode<XmlExp>( "endFrame" ) )->SetData( ToString( m_aniInfo.endFrame ) );
	pXmlNode->AttachChild( CreateXmlNode<XmlExp>( "tickPerFrame" ) )->SetData( ToString( m_aniInfo.tickPerFrame ) );
	pXmlNode->AttachChild( CreateXmlNode<XmlExp>( "fps" ) )->SetData( ToString( m_aniInfo.fps ) );
}
//-----------------------------------------------------------------------------------
void Export::BuildINode( INode* pNode, XmlExp* pParentXML )
{
	if(!pNode)
		return;

	XmlExp* pXmlNode = WriteXMLObject(pNode, pParentXML );
	if( pXmlNode == NULL )
	{
		pXmlNode = WriteXmlNode( pNode, "Node" );
		pXmlNode->AddAttribute( "name", pNode->GetName() );
		if( m_pXmlRoot == NULL )
			pParentXML = m_pXmlRoot = pXmlNode;
		else
			pParentXML->AttachChild( pXmlNode );
	}

	pXmlNode->SetINode( pNode );
	WriteAnimation( pNode, pXmlNode );
	int iChild = pNode->NumberOfChildren();
	for(int i=0; i<iChild; ++i)
	{
		INode* pChild = pNode->GetChildNode(i);

		if(pChild)
		{
			BuildINode(pChild, pXmlNode );
		}
	}
}
//-----------------------------------------------------------------------------------
XmlExp* Export::WriteXmlNode( INode* pNode, char* szName )
{
	XmlExp* pXmlNode = CreateXmlNode<XmlExp>( szName );
	INode* pParent = pNode->GetParentNode();

	Matrix3 matLocal;
	Matrix3 matWorld = pNode->GetObjTMAfterWSM( 0 );

	if( pParent == NULL )
	{
		matLocal = matWorld;
	}
	else
	{
		Matrix3 matParent = pParent->GetObjTMAfterWSM( 0 );
		matLocal = matWorld * Inverse( matParent );
	}


	Matrix3 maxMat = matLocal;
	if( GetLeftHand() )
		RightToLeft( &maxMat, &matLocal );

	base::Matrix mat;	mat.Init();		MaxToBaseMatrix( &mat, &maxMat );

	XmlExp* pTransform = (XmlExp*)pXmlNode->AttachChild( CreateXmlNode<XmlExp>( "Transform" ) );
	pTransform->AttachChild( CreateXmlNode<XmlExp>( "LocalMatrix" ) )->SetData(  ToString( mat ) );
	return pXmlNode;
}
//-----------------------------------------------------------------------------------
void Export::MaxToBaseMatrix( base::Matrix*  dest, Matrix3* pSrc )
{
	Point3	v1;
	
	v1	= pSrc->GetRow(0);	dest->_11 = v1.x;	dest->_12 = v1.y;	dest->_13 = v1.z;
	v1	= pSrc->GetRow(1);	dest->_21 = v1.x;	dest->_22 = v1.y;	dest->_23 = v1.z;
	v1	= pSrc->GetRow(2);	dest->_31 = v1.x;	dest->_32 = v1.y;	dest->_33 = v1.z;
	v1  = pSrc->GetRow(3);	dest->_41 = v1.x;	dest->_42 = v1.y;	dest->_43 = v1.z;
}
//-----------------------------------------------------------------------------------
void Export::RightToLeft(Matrix3* pDst, Matrix3* pSrc )
{
	Point3	v1, v2;
	
	v1	= pSrc->GetRow(0);	v2.x = v1.x;	v2.y = v1.z;	v2.z = v1.y;   	pDst->SetRow( 0, v2 );
	v1	= pSrc->GetRow(2);	v2.x = v1.x;	v2.y = v1.z;	v2.z = v1.y;    pDst->SetRow( 1, v2 );
	v1	= pSrc->GetRow(1);	v2.x = v1.x;	v2.y = v1.z;	v2.z = v1.y;    pDst->SetRow( 2, v2 );
	v1	= pSrc->GetRow(3);	v2.x = v1.x;	v2.y = v1.z;	v2.z = v1.y;    pDst->SetRow( 3, v2 );
}
//-----------------------------------------------------------------------------------
bool Export::ExportXML( const char* szFileName )
{
	WriteXml( szFileName, m_pXmlRoot );
	return true;
}
//-----------------------------------------------------------------------------------
void Export::GetAnimationInfo()
{
	int		 iTick = GetTicksPerFrame();
	Interval range = m_pInterface->GetAnimRange();

	m_aniInfo.beginFrame = range.Start() / iTick;
	m_aniInfo.endFrame = range.End() / iTick;
	m_aniInfo.fps = GetFrameRate();
	m_aniInfo.tickPerFrame = iTick;
}
//-----------------------------------------------------------------------------------
bool Export::IsBone( INode* pNode )
{
	if(pNode == NULL)
		return false;

	ObjectState  objectState = pNode->EvalWorldState(0);
	if(objectState.obj == NULL)
		return false;
	if(objectState.obj->ClassID() == BONE_OBJ_CLASSID)   
		return true;
	//include with Helper
	if(objectState.obj->ClassID() == Class_ID(BONE_CLASS_ID, 0)) 
		return true;
	//include with Helper
	if(objectState.obj->ClassID() == Class_ID(DUMMY_CLASS_ID, 0)) 
		return false;
 
	Control* control = pNode->GetTMController();
	Class_ID cid = control->ClassID(); 
	//others biped parts || biped root 
	if( cid == BIPSLAVE_CONTROL_CLASS_ID || cid == BIPBODY_CONTROL_CLASS_ID )  
		return true;

	return false;
}
//-----------------------------------------------------------------------------------
