//XmlExp.cpp
#include "Max.h"
#include "iparamb2.h"
#include "iskin.h"
#include "CS/Phyexp.h"
#include "modstack.h"

#include "XmlExp.h"
#include "./xmlnode/StringUtil.h"

int XmlExp::m_nRef = 0;
XmlExp::INodeIndexMap XmlExp::m_indexMap;

//-----------------------------------------------------------------------------------
Modifier* FindModifier( INode* pNode, const Class_ID& classID )
{
	if( pNode == NULL )
		return 0;

	Object* pObj = pNode->GetObjectRef();
	if( pObj == NULL )
		return NULL;

	while( pObj->SuperClassID() == GEN_DERIVOB_CLASS_ID )
	{
		IDerivedObject* pDerivedObj = (IDerivedObject*)pObj;

		int index = 0;
		while( index < pDerivedObj->NumModifiers() )
		{
			Modifier* pModifier = pDerivedObj->GetModifier( index );
			if( pModifier->ClassID() == classID )
				return pModifier;
			++index;
		}

		pObj = pDerivedObj->GetObjRef();
	}

	return 0;
}
//-----------------------------------------------------------------------------------
int XmlExp::FindID( INode* pINode )
{
	INodeIndexMap::iterator	iter = m_indexMap.find( pINode );
	if( iter != m_indexMap.end() )
	{
		return iter->second;
	}

	return -1;
}
//-----------------------------------------------------------------------------------
XmlExp::XmlExp()
{
	m_pINode = NULL;
	m_nIndex = -1;
}
//-----------------------------------------------------------------------------
XmlExp::~XmlExp()
{
}
//-----------------------------------------------------------------------------
const char*	XmlExp::GetClassName()
{
	return "XmlExp";
}
//-----------------------------------------------------------------------------
void XmlExp::SetINode( INode* pINode )
{
	assert( m_pINode == NULL );
	m_nIndex = m_nRef;
	m_indexMap.insert( std::make_pair( pINode, m_nIndex ) );
	AddAttribute( "index", ToString( m_nIndex ).c_str() );
	++m_nRef;
	m_pINode = pINode;
}
//-----------------------------------------------------------------------------
void XmlExp::BuildBoneWeight()
{
	if( m_pINode )
	{
		Modifier*	pMod = (Modifier*)FindModifier(m_pINode, Class_ID(PHYSIQUE_CLASS_ID_A, PHYSIQUE_CLASS_ID_B) );
		if(pMod)
		{
			WritePhysiqueWeight( pMod );
			return;
		}

		pMod = (Modifier*)FindModifier(m_pINode, SKIN_CLASSID);
		if(pMod)
		{
			WriteSkinWeight( pMod );
			return;
		}
	}

	XmlNode::ChildList* pChildList = &m_child;
	XmlNode::ChildList::iterator iter = pChildList->begin();
	XmlNode::ChildList::iterator endIter = pChildList->end();
	for( ; iter != endIter; ++iter )
	{
		XmlExp* pChildXml = (XmlExp*)*iter;
		pChildXml->BuildBoneWeight();
	}
}
//-----------------------------------------------------------------------------
void XmlExp::WritePhysiqueWeight( Modifier* pMod )
{
	INode*				pINode = m_pINode;
	IPhysiqueExport*	pExport = NULL;
	IPhyContextExport*	pContext = NULL;

	pExport = (IPhysiqueExport*)pMod->GetInterface(I_PHYINTERFACE);		
	if(pExport == NULL)
		return;

	pContext = (IPhyContextExport*)pExport->GetContextInterface(pINode);	
	if(pContext == NULL)
	{
		pMod->ReleaseInterface(I_PHYINTERFACE, pExport);
		return;
	}

	pContext->ConvertToRigid(TRUE);					// Context ؼ   Rigid ٲ۴.
	pContext->AllowBlending(TRUE);					// Blending Ȱȭ

	int nVtx = pContext->GetNumberVertices();		// Bone ޴  rot
	if(nVtx <= 0)
		return;

	XmlExp* pRoot = (XmlExp*)GetRoot();
	BlendVector* pBlend = new BlendVector[nVtx];
	for(int j=0; j< nVtx; ++j)
	{
		IPhyVertexExport*		pPhyVtxExport = (IPhyVertexExport*)pContext->GetVertexInterface(j);
		IPhyBlendedRigidVertex* pPhyBlend	= (IPhyBlendedRigidVertex*)pPhyVtxExport;

		if(!pPhyVtxExport)
			continue;

		if(RIGID_TYPE == pPhyVtxExport->GetVertexType())
		{
			// ε Ѵ.
			INode*	pBone	= ((IPhyRigidVertex*)pPhyVtxExport)->GetNode();
			INT		nBone	= pRoot->FindID(pBone);
			FLOAT	fWgt	= 1.f;		//RIGID_TYPE Weight=1 ̴.
			
			if( nBone >= 0 )
			{
				BlendData blend = { nBone, fWgt };
				pBlend[j].push_back( blend );
			}
			continue;
		}

		int numBones = pPhyBlend->GetNumberNodes();
		for(int k = 0; k< numBones; ++k)
		{
			// k°  ã´.
			INode*	pBone	= pPhyBlend->GetNode(k);
			INT		nBone	= pRoot->FindID(pBone);
			FLOAT	fWgt	= pPhyBlend->GetWeight(k);
			
			if( nBone < 0 )
				continue;

			//   
			if(fWgt<0.00005f)
				continue;
		
			BlendData blend = { nBone, fWgt };
			pBlend[j].push_back( blend );
		}
	}

	WriteRigging( nVtx, pBlend, "physique" );
	
	pExport->ReleaseContextInterface(pContext);
	pMod->ReleaseInterface(I_PHYINTERFACE, pExport);
	delete [] pBlend;
}
//-----------------------------------------------------------------------------
void XmlExp::WriteSkinWeight( Modifier* pMod )
{
	INode*				pINode = m_pINode;
	ISkin*				pExport = NULL;
	ISkinContextData*	pContext = NULL;

	pExport = (ISkin*)pMod->GetInterface(I_SKIN);		// Skin ü
	if(!pExport)
		return;

	pContext = pExport->GetContextInterface(pINode);		// SKin Export  Context Ѵ.
	if(!pContext)
		return;

	int nVtx = pContext->GetNumPoints();				// Bone ޴  

	if(nVtx <=0)
		return;

	XmlExp* pRoot = (XmlExp*)GetRoot();
	BlendVector* pBlend = new BlendVector[nVtx];

	for(int j=0; j<nVtx;  ++j)
	{
		int numBones = pContext->GetNumAssignedBones(j);
		
		for(int k=0; k<numBones; ++k)
		{
			int assignedBone = pContext->GetAssignedBone(j, k);

			if(assignedBone < 0)
				continue;

			INode*	pBone	= pExport->GetBone(assignedBone);
			INT		nBone	= pRoot->FindID(pBone);
			FLOAT	fWgt	= pContext->GetBoneWeight(j, k);

			if( nBone < 0 )
				continue;
			//   
			if(fWgt < 0.00005f)
				continue;

			BlendData blend = { nBone, fWgt };
			pBlend[j].push_back( blend );
		}
	}

	WriteRigging( nVtx, pBlend, "skin" );
	pMod->ReleaseInterface(I_SKIN, pExport );
	delete [] pBlend;
}
//-----------------------------------------------------------------------------
void XmlExp::WriteRigging( int nVertexNum, BlendVector* pBlend, const char* riggingType )
{
	XmlExp* pMesh = (XmlExp*)this->FindFirstChild( "mesh" );
	if( pMesh == NULL )
	{
		//  ü ° ƽ ÷ζ Ȥ !!!
		pMesh = CreateXmlNode<XmlExp>( "mesh" );
		this->AttachChild( pMesh );
	}
	pMesh->AddAttribute( "riggingType", riggingType );

	XmlExp* pIndexNode = CreateXmlNode<XmlExp>( "boneIndex" );
	pMesh->AttachChild( pIndexNode );
	pIndexNode->AddAttribute( "count", ToString((int)nVertexNum ).c_str() );		
	
	//Bone index 
	for( int i = 0; i < nVertexNum; ++i )
	{
		BlendVector& blend = pBlend[i];
		std::vector<int> refArray;
		for( int j = 0; j < blend.size(); ++j )
			refArray.push_back( blend[j].index );

		XmlExp* pData = CreateXmlNode<XmlExp>( "data" );
		pIndexNode->AttachChild( pData );
		pData->SetData( ToString( refArray ) );
	}

	XmlExp* pWeightNode = CreateXmlNode<XmlExp>( "weight" );
	pMesh->AttachChild( pWeightNode );
	pWeightNode->AddAttribute( "count", ToString((int)nVertexNum ).c_str() );
	
	//Weight 
	for( int i = 0; i < nVertexNum; ++i )
	{
		BlendVector& blend = pBlend[i];
		std::vector<float> floatVector;
		for( int j = 0; j < blend.size(); ++j )
			floatVector.push_back( blend[j].weight );

		XmlExp* pData = CreateXmlNode<XmlExp>( "data" );
		pWeightNode->AttachChild( pData );
		pData->SetData( ToString( floatVector ) );
	}
}
